{-# LANGUAGE ViewPatterns #-}
module Builtins
(getBuiltins
) where

import CoreSyn
import DataCon
import FastString
import GHC hiding (compileExpr, exprType)
import HscMain (compileExpr)
import HscTypes
import Id
import Linker (extendLinkEnv)
import MkCore
import TysWiredIn
import UniqSupply
import Unique
import Var

getBuiltins :: UniqSupply -> Ghc [Id]
getBuiltins us = do
  let (take 4 -> [um1, u0, u1, uca]) = uniqsFromSupply us

  ints <- mapM (uncurry mkInt) [((-1), um1), (0, u0), (1, u1)]
  let listNil  = dataConWrapId nilDataCon
  let listCons = dataConWrapId consDataCon
  charA <- mkChar 'a' uca

  return $ ints ++ [listNil, listCons, charA]

mkInt :: Int -> Unique -> Ghc Id
mkInt i u = do
  let letId = mkSysLocal (mkFastString (show i)) u intTy
  let rhs = mkIntExprInt i
  let codeToRun = mkCoreLet (NonRec letId rhs) (Var letId)

  hval <- withSession $ \hsc -> liftIO $ compileExpr hsc noSrcSpan codeToRun 
  {- builds: let letId = I# i in letId -}
  liftIO $ extendLinkEnv [(varName letId, hval)]
  return letId

mkChar :: Char -> Unique -> Ghc Id
mkChar c u = do
  let letId = mkSysLocal (mkFastString (show c)) u charTy
  let rhs = mkCharExpr c
  let codeToRun = mkCoreLet (NonRec letId rhs) (Var letId)

  hval <- withSession $ \hsc -> liftIO $ compileExpr hsc noSrcSpan codeToRun 
  {- builds: let letId = C# c in letId -}
  liftIO $ extendLinkEnv [(varName letId, hval)]
  return letId
