{-# LANGUAGE BangPatterns #-}
module Engine.EngineHValue 
( EngineHValue
, toHValue
, toStableName
, toType
, mkEngineHValue
, hashEngineHValue
, getConstructor
) where

import ByteCodeLink
import CoreMonad
import Data.Array
import DataCon
import GHC
import HscTypes
import Linker
import RtClosureInspect
import System.Mem.StableName
import TcEnv
import Type
import Util.GHC(initTcForLookup)
import Control.Applicative
import Debug.Trace

data EngineHValue = EngineHValue { toHValue     :: !HValue
                                 , toStableName :: !(StableName HValue)
                                 , toType       :: !Type
                                 }

instance Eq EngineHValue where
  hv1 == hv2 = toStableName hv1 == toStableName hv2 
            && toType hv1 == toType hv2
          

mkEngineHValue :: HValue -> Type -> IO EngineHValue
mkEngineHValue !arg !ty = do
  snArg <- makeStableName $! arg
  return $ EngineHValue { toHValue     = arg
                        , toStableName = snArg
                        , toType       = ty
                        }
hashEngineHValue :: EngineHValue -> Int
hashEngineHValue = hashStableName . toStableName

getConstructor :: (GhcMonad m) => EngineHValue -> m (Maybe DataCon)
getConstructor ehv = do
  chaseConstructor $! toHValue ehv

chaseConstructor :: (GhcMonad m) => HValue -> m (Maybe DataCon)
chaseConstructor !hv = do
  closure <- liftIO $ getClosureData hv
  case tipe closure  of
    Indirection _ -> chaseConstructor (ptrs closure ! 0)
    Constr -> do
      withSession $ \hscEnv -> liftIO $ initTcForLookup hscEnv $ do
        eDcname <- dataConInfoPtrToName (infoPtr closure)
        case eDcname of
          Left _       -> return Nothing
          Right dcName -> Just <$> tcLookupDataCon dcName
    _ -> return Nothing
