{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE TemplateHaskell #-}
module FrontEnd.Options where

import Data.Int
import Distribution.Simple.Setup
import Plan (Depth)
import Engine (Strategy, fifoStrategy)
import IdDB (IdSource(..))
import Util.FCLabels
import Data.Monoid

data RuntimePlan
  = DepthBounded
  | RandomTree
  deriving Eq

data IrulanOptions
  = IrulanOptions { runtimePlanType        :: !(Flag RuntimePlan)
                  , maximumDepth           :: !(Flag Depth) --          { for depth bounded plans }
                  , maxRuntimePerFunction  :: !(Flag Int)   -- seconds  { for Tree-like plans }
                  , maxExpressionTime      :: !(Flag Int)   -- microseconds (1/10^6)
                  , maxExpressionAllocs    :: !(Flag Int64) -- bytes
                  , outputExperimentTrace  :: !(Flag Bool)
                  , outputDebugInformation :: !(Flag Bool)
                  , outputStatistics       :: !(Flag FilePath)
                  , useSimpleOutput        :: !(Flag Bool)
                  , useCaseStatement       :: !(Flag Bool)
                  , sourceIdsFrom          :: !(Flag IdSource)
                  , explicitIdSource       :: !([String])
                  , cache                  :: !(Last (Maybe Strategy))
                  , quiet                  :: !(Flag Bool)
                  }

mkLabels [''IrulanOptions]



defaultOptions :: IrulanOptions
defaultOptions
  = IrulanOptions { runtimePlanType        = Flag $ RandomTree
                  , maximumDepth           = Flag $ 5
                  , maxRuntimePerFunction  = Flag $ 10
                  , maxExpressionTime      = Flag $ 1 * (10^ (6 :: Integer)) -- 1 seconds
                  , maxExpressionAllocs    = Flag $ 1024 * 1024 * 128 -- 128Meg
                  , outputExperimentTrace  = Flag $ False
                  , outputDebugInformation = Flag $ False
                  , outputStatistics       = NoFlag
                  , useSimpleOutput        = Flag $ True
                  , useCaseStatement       = Flag $ True
                  , sourceIdsFrom          = Flag $ SourceModule
                  , explicitIdSource       = []
                  , cache                  = Last . Just $ Just (fifoStrategy 1000)
                  , quiet                  = Flag $ False
                  }


instance Monoid IrulanOptions where
  mempty 
    = IrulanOptions { runtimePlanType        = mempty
                    , maximumDepth           = mempty
                    , maxRuntimePerFunction  = mempty
                    , maxExpressionTime      = mempty
                    , maxExpressionAllocs    = mempty
                    , outputExperimentTrace  = mempty
                    , outputDebugInformation = mempty
                    , outputStatistics       = mempty
                    , useSimpleOutput        = mempty
                    , useCaseStatement       = mempty
                    , sourceIdsFrom          = mempty
                    , explicitIdSource       = mempty
                    , cache                  = mempty
                    , quiet                  = mempty
                    }

  mappend a b 
    = IrulanOptions { runtimePlanType        = combine runtimePlanType
                    , maximumDepth           = combine maximumDepth
                    , maxRuntimePerFunction  = combine maxRuntimePerFunction
                    , maxExpressionTime      = combine maxExpressionTime
                    , maxExpressionAllocs    = combine maxExpressionAllocs
                    , outputExperimentTrace  = combine outputExperimentTrace
                    , outputDebugInformation = combine outputDebugInformation
                    , outputStatistics       = combine outputStatistics
                    , useSimpleOutput        = combine useSimpleOutput
                    , useCaseStatement       = combine useCaseStatement
                    , sourceIdsFrom          = combine sourceIdsFrom
                    , explicitIdSource       = combine explicitIdSource
                    , cache                  = combine cache
                    , quiet                  = combine quiet
                    }
      where
        combine f = f a `mappend` f b
