{-# LANGUAGE ViewPatterns #-}
module Plan.Debug
( dotifyTreePlan
) where

import Control.Applicative hiding (empty)
import Control.Monad.State
import Data.Graph.Inductive hiding (mkNode, mkEdge)
import Data.GraphViz
import Format
import Plan.Experiment
import Plan.TreePlan
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import Util
import Util.Graph
import qualified Util.UniqMap as UniqMap
import Util.UniqMap (UniqMap)

dotifyTreePlan :: Int -> TreePlan -> String
dotifyTreePlan depth = printDotGraph . (\gr -> graphToDot True gr [] mkNode mkEdge) . minimizeGraph . graphTreePlan depth

mkNode :: (Format t) => (a, Maybe t) -> [Attribute]
mkNode (_, Nothing) = [Style [SItem Filled []], FillColor (RGB 0 0 0)]
mkNode (_, Just ex) = mkLabel (escapeNL $ format ex)

mkEdge :: (a,b, ExperimentResult) -> [Attribute]
mkEdge (_,_, Ok       ) = mkLabel "."
mkEdge (_,_, Bottoms x) = mkLabel $ "?" ++ (escapeNL $ format x)

mkLabel :: EscString -> [Attribute]
mkLabel l = [Label (StrLabel l)]

escapeNL :: String -> String
escapeNL = map (\x -> if x == '\n' then ' ' else x)

doParent :: Node -> Maybe (Node, ExperimentResult) -> Gr (Maybe Experiment) ExperimentResult 
                                                   -> Gr (Maybe Experiment) ExperimentResult
doParent _  Nothing              = id
doParent nn (Just (preNode, er)) = insEdge (preNode, nn, er)


graphTreePlan :: Int -> TreePlan -> Gr (Maybe Experiment) ExperimentResult
graphTreePlan depth plan = execState (graphTreePlan' Nothing depth plan) empty

graphTreePlan' :: (Maybe (Node, ExperimentResult)) -> Int -> TreePlan -> State (Gr (Maybe Experiment) ExperimentResult) ()
graphTreePlan' _ 0 _ = return ()
graphTreePlan' parent n TPSplit { tpChildren = kids } 
  | Seq.null kids = return ()
  | otherwise     = do
    F.sequence_ $ graphTreePlan' parent n <$> kids
    return ()
graphTreePlan' parent n TPStep { tpExperiment = experiment
                               , tpOnOk       = onOk
                               , tpOnBottom   = onBottom
                               } = do
  [nn] <- gets $ newNodes 1
  modify (doParent nn parent . insNode (nn, Just experiment))
  case onOk of
    Left tp  -> do graphTreePlan' (Just (nn, Ok)) (n-1) tp
    Right mp -> forM_ (UniqMap.elems mp) $ \(dc, elts) -> do
      F.forM_ elts $ \elt -> graphTreePlan' (Just (nn, Ok)) (n-1) elt
  
  F.sequence_ . flip Map.mapWithKey onBottom $ \botId plans -> do
    F.forM_ plans $ graphTreePlan' (Just (nn, Bottoms botId)) (n-1)
  return ()
