{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE PatternGuards #-}
{-# OPTIONS_GHC -fno-warn-overlapping-patterns #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE ViewPatterns #-}
module Plan.IExperiment
( IExperiment (iExperimentRoot, iExperimentArgs)
, Depth
, WeakIExperiment (WeakIExperiment)
, makeInitialExperiment
, iExperimentToExperiment
, iExperimentCost
, iExperimentType
, IExperimentArg (..)
, IExperimentRoot(..)
, addBottomArg
, substExpFn
, isSimpleCtrExperiment
, substBottomId
, bottomIdTypes
, destructIExperiment
) where

import Bottom
import Control.Applicative
import DataCon
import Data.List
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Data.Sequence (Seq, (|>), (><))
import Format
import Id
import Name
import Plan.Experiment
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Data.Traversable as T
import Type
import TysPrim (anyPrimTy)
import TysWiredIn (boolTy)
import UniqSet (uniqSetToList)
import Unique
import Util.GHC
import Var
import Debug.Trace

newtype Depth = Depth Integer
  deriving (Eq, Ord, Show, Num, Enum, Read)

instance Format Depth where
  format (Depth i) = show i

data IExperiment
  = IExperiment { iExperimentRoot  :: !IExperimentRoot
                , iExperimentArgs  :: !(Seq IExperimentArg)
                , iExperimentSubst :: !TvSubst
                }

data IExperimentRoot
  = IIdRoot !Id
  | IDestructRoot
  { iDestructCaseOf      :: !IExperiment
  , iDestructConstructor :: !DataCon
  , iDestructVarIndex    :: !Int
  , iDestructResultType  :: !Type
  }

data IExperimentArg
  = IArgExperiment    !IExperiment
  | IArgBottom        !Type  !BottomId
  | IArgUnconstrained !TyVar !BottomId -- BottomId to use if we need to convert this
                                       -- into a ArgBottom

newtype WeakIExperiment = WeakIExperiment IExperiment
newtype WeakIExperimentArg = WeakIExperimentArg IExperimentArg
newtype WeakIExperimentRoot = WeakIExperimentRoot IExperimentRoot

instance Eq WeakIExperiment where
  (WeakIExperiment l) == (WeakIExperiment r) 
    =  (WeakIExperimentRoot . iExperimentRoot $ l) == (WeakIExperimentRoot . iExperimentRoot $ r) && 
       Seq.length (iExperimentArgs l) == Seq.length (iExperimentArgs r) && 
       and (zipWith (==) (F.toList $ WeakIExperimentArg <$> iExperimentArgs l) 
                         (F.toList $ WeakIExperimentArg <$> iExperimentArgs r))

instance Ord WeakIExperiment where
  (WeakIExperiment l) `compare` (WeakIExperiment r) = mkRep l `compare` mkRep r
    where
      mkRep :: IExperiment -> (WeakIExperimentRoot, Seq WeakIExperimentArg)
      mkRep exp = ((WeakIExperimentRoot . iExperimentRoot) exp, WeakIExperimentArg <$> iExperimentArgs exp)

instance Eq WeakIExperimentArg where
  (WeakIExperimentArg l) == (WeakIExperimentArg r) = l =~= r
    where
      (IArgExperiment sl) =~= (IArgExperiment sr) = (WeakIExperiment sl) == (WeakIExperiment sr)
      (IArgBottom _ bl) =~= (IArgBottom _ br) = bl == br
      (IArgUnconstrained _ bl) =~= (IArgUnconstrained _ br) = bl == br

instance Ord WeakIExperimentArg where
  (WeakIExperimentArg l) `compare` (WeakIExperimentArg r) = mkRep l `compare` mkRep r
    where
      mkRep :: IExperimentArg -> Either WeakIExperiment (Either BottomId BottomId)
      mkRep (IArgExperiment sub)      = Left (WeakIExperiment sub)
      mkRep (IArgBottom _ bid)        = Right (Left bid)
      mkRep (IArgUnconstrained _ bid) = Right (Right bid)

instance Eq WeakIExperimentRoot where
  (WeakIExperimentRoot (IIdRoot l)) == (WeakIExperimentRoot (IIdRoot r)) = l == r
  (WeakIExperimentRoot (l@IDestructRoot {})) == (WeakIExperimentRoot (r@IDestructRoot {}))
    = and [ iDestructConstructor l == iDestructConstructor r
          , iDestructVarIndex l    == iDestructVarIndex r 
          , WeakIExperiment (iDestructCaseOf l) == WeakIExperiment (iDestructCaseOf r)
          ] 
  _ == _ = False
  

instance Ord WeakIExperimentRoot where
  (WeakIExperimentRoot l) `compare` (WeakIExperimentRoot r) = mkRep l `compare` mkRep r
    where
      mkRep :: IExperimentRoot -> Either Id (DataCon, Int, WeakIExperiment)
      mkRep (IIdRoot x) = Left x
      mkRep (d@IDestructRoot {}) = Right (iDestructConstructor d, iDestructVarIndex d, WeakIExperiment $ iDestructCaseOf d)
  

instance Format IExperimentArg where
  format (IArgExperiment e)
    | Seq.null (iExperimentArgs e) = format e
    | otherwise                       = "(" ++ format e ++ ")"
  format (IArgBottom t b) = "?<" ++ show t ++ ">" ++ format b
  format (IArgUnconstrained t b) = "*<" ++ show t ++ show b ++ ">"

instance Format IExperiment where
  format exp = unwords $ [ format (iExperimentRoot exp)
                         , show $ iExperimentSubst exp 
                         ] ++ (F.toList $ format <$> iExperimentArgs exp)

instance Format IExperimentRoot where
  format (IIdRoot x) = show x
  format (IDestructRoot { iDestructCaseOf
                        , iDestructConstructor
                        , iDestructVarIndex
                        , iDestructResultType
                        }) = "case (" ++ format iDestructCaseOf ++ ") of " ++ show iDestructConstructor ++ " -> " ++ show iDestructVarIndex

iExperimentCost :: IExperiment -> Depth
iExperimentCost (iExperimentArgs -> args)           = 1 + (getSum $ F.foldMap (Sum . cost) args)
  where
    cost (IArgBottom _ _)                           = 1
    cost (IArgExperiment (iExperimentCost -> e))    = e
    cost (IArgUnconstrained _ _)                    = 1

iExperimentRootType :: IExperimentRoot -> Type
iExperimentRootType (IIdRoot x) 
  = varType x
iExperimentRootType (IDestructRoot { iDestructResultType })
  = iDestructResultType

iExperimentType :: IExperiment -> Type
iExperimentType IExperiment { iExperimentRoot , iExperimentArgs , iExperimentSubst } = resultType
  where
    initType = snd . splitForAllTys $ iExperimentRootType iExperimentRoot

    resultType = F.foldl reduceArg (substTy iExperimentSubst initType) iExperimentArgs

    reduceArg :: Type -> IExperimentArg -> Type
    reduceArg t _
      | isFunTy t = funResultTy t
    reduceArg t arg = error $ "reduceArg: " ++ (format arg) ++ show t ++ " - " ++ format iExperimentRoot

iExperimentToExperiment :: IExperiment -> Experiment
iExperimentToExperiment (iexp@IExperiment { iExperimentRoot, iExperimentArgs })
  = let outExp = Experiment { experimentRoot       = mkExpRoot iExperimentRoot
                            , experimentArgs       = experimentTyArgs >< experimentValArgs
                            , experimentIsProperty = isProperty
                            , experimentHasBottom  = hasBottomArg iexp
                            }
    in outExp
  where
    isProperty = isPropertyRoot
              && iExperimentType iexp == boolTy

        
    (experimentTyArgs, experimentRawType) = splitIExpForalls iexp
    (_ ,experimentValArgs) = T.mapAccumL mkExpArguments experimentRawType iExperimentArgs

    mkExpRoot (IIdRoot i) = IdRoot i
    mkExpRoot (IDestructRoot { iDestructCaseOf , iDestructConstructor , iDestructVarIndex , iDestructResultType }) 
      = DestructRoot { destructCaseOf = iExperimentToExperiment iDestructCaseOf
                     , destructConstructor = iDestructConstructor
                     , destructVarIndex = iDestructVarIndex
                     , destructResultType = iDestructResultType
                     }

    isPropertyRoot
      | IIdRoot x <- iExperimentRoot = "prop_" `isPrefixOf` (occNameString . getOccName $ x)
      | otherwise   = False

    mkExpArguments :: Type -> IExperimentArg -> (Type, ExperimentArg)
    mkExpArguments t (IArgExperiment e)
      | isFunTy t = (funResultTy t, ArgExperiment e')
      where
        e' = iExperimentToExperiment e
    mkExpArguments t (IArgBottom bt bid)
      | isFunTy t = (funResultTy t, ArgBottom bt bid)
    mkExpArguments t (IArgUnconstrained _ _)
      | isFunTy t = (funResultTy t, ArgUnconstrained)
    mkExpArguments t unkownArg
      = error $ "mkExpArguments: " ++ show t ++ " " ++ "; " ++ format unkownArg ++ " ;" ++ format iExperimentRoot


splitIExpForalls :: IExperiment -> (Seq ExperimentArg, Type)
splitIExpForalls IExperiment { iExperimentRoot , iExperimentSubst } 
  = (experimentTypeArgs, outTy)
  where
    (freeVars, rawType) = splitForAllTys . iExperimentRootType $ iExperimentRoot 

    outTyArgs = (\fv -> fromMaybe (mkTyVarTy fv) (lookupTyVar iExperimentSubst fv)) 
            <$> freeVars

    experimentTypeArgs = Seq.fromList $ (ArgType . anyizeFreeVars) <$> outTyArgs
    outTy = substTy iExperimentSubst rawType

anyizeFreeVars :: Type -> Type
anyizeFreeVars ty = substTyWith freeVars tarTys ty
  where
    freeVars = uniqSetToList . tyVarsOfType $ ty
    tarTys   = map (const anyPrimTy) freeVars


makeInitialExperiment :: Id -> TvSubst -> IExperiment
makeInitialExperiment sourceId subst
  = IExperiment { iExperimentRoot       = IIdRoot sourceId
                , iExperimentArgs       = Seq.empty
                , iExperimentSubst      = subst `composeTvSubsts` baseSubst
                }
  where
    (tyVars, _) = splitForAllTys . varType $ sourceId
    baseSubst = zipOpenTvSubst tyVars (mkTyVarTys tyVars)

substExpFn :: TvSubst -> IExperiment -> IExperiment
substExpFn subst exp 
  = exp { iExperimentArgs     = newIExperimentArgs
        , iExperimentSubst    = subst'
        , iExperimentRoot     = newRoot
        }
  where
    subst' = composeTvSubsts subst (iExperimentSubst exp)

    newIExperimentArgs = (substArg subst') <$> iExperimentArgs exp

    newRoot 
      | r@(IIdRoot _)      <- iExperimentRoot exp = r
      | r@IDestructRoot {} <- iExperimentRoot exp = r { iDestructCaseOf     = substExpFn subst' (iDestructCaseOf r)
                                                      , iDestructResultType = substTy subst' (iDestructResultType r)
                                                      }

    substArg :: TvSubst -> IExperimentArg -> IExperimentArg
    substArg subst (IArgExperiment exp) = IArgExperiment (substExpFn subst exp)
    substArg subst (IArgBottom ty bid)  = IArgBottom (substTy subst ty) bid
    substArg subst (IArgUnconstrained tv bid)
      | isTyVarTy ty  = IArgUnconstrained (getTyVar "substArg" ty) bid
      | otherwise     = IArgBottom ty bid
      where
        ty = substTy subst (mkTyVarTy tv)

addBottomArg :: IExperiment -> BottomId -> IExperiment
addBottomArg exp bid 
  | isFunTy eType = exp { iExperimentArgs = iExperimentArgs exp |> IArgBottom bty bid }
  | otherwise  = error $ "addBottomArg: " ++ format exp ++ "; " ++ show bid ++ "; " ++ show eType
  where
    eType = iExperimentType exp
    (bty,_) = splitFunTy eType

isSimpleCtrExperiment :: IExperiment -> Bool
isSimpleCtrExperiment iexp = isSimpleCtrExperimentRoot
                           && (F.all isBottomArg . iExperimentArgs $ iexp)
  where
    isSimpleCtrExperimentRoot
      | IIdRoot x <- (iExperimentRoot iexp) = isJust . isDataConId_maybe $ x
      | otherwise = False
                  
isBottomArg :: IExperimentArg -> Bool
isBottomArg (IArgBottom _ _) = True
isBottomArg _                = False

substBottomId :: BottomId -> IExperimentArg -> IExperiment -> IExperiment
substBottomId bid arg e@IExperiment { iExperimentRoot, iExperimentArgs }
  = e { iExperimentArgs = substBottomId' <$> iExperimentArgs
      , iExperimentRoot = substBottomId'' iExperimentRoot 
      }
  where
    substBottomId' :: IExperimentArg -> IExperimentArg
    substBottomId' b@(IArgBottom _ bid')
      | bid == bid'                          = arg
      | otherwise                            = b
    substBottomId' (IArgExperiment exp)      
      = IArgExperiment (substBottomId bid arg exp)
    substBottomId' a@(IArgUnconstrained _ _) = a
    
    substBottomId'' i@(IIdRoot _) = i
    substBottomId'' i@(IDestructRoot { iDestructCaseOf }) 
      = i { iDestructCaseOf = substBottomId bid arg iDestructCaseOf }

hasBottomArg :: IExperiment -> Bool
hasBottomArg IExperiment { iExperimentRoot
                         , iExperimentArgs
                         } = getAny (F.foldMap (Any . isBottomArg) iExperimentArgs `mappend` (hasBottomRoot iExperimentRoot))
  where
    hasBottomRoot (IIdRoot _) = Any False
    hasBottomRoot (IDestructRoot { iDestructCaseOf }) = Any $ hasBottomArg iDestructCaseOf

bottomIdTypes :: IExperiment -> Map BottomId Type
bottomIdTypes IExperiment { iExperimentRoot , iExperimentArgs } = (F.foldMap bottomIdTypes' iExperimentArgs) 
                                                                  `mappend` bottomIdTypesR
  where
    bottomIdTypesR = case iExperimentRoot of
      IIdRoot _                         -> Map.empty
      IDestructRoot { iDestructCaseOf } -> bottomIdTypes iDestructCaseOf


    bottomIdTypes' :: IExperimentArg -> Map BottomId Type
    bottomIdTypes' (IArgBottom t bid)      = Map.singleton bid t
    bottomIdTypes' (IArgExperiment exp)    = bottomIdTypes exp
    bottomIdTypes' (IArgUnconstrained _ _) = Map.empty


destructIExperiment :: IExperiment -> DataCon -> Int -> Type -> IExperiment
destructIExperiment orig dc varIdx varType
  = IExperiment { iExperimentRoot  = IDestructRoot { iDestructCaseOf      = orig
                                                   , iDestructConstructor = dc
                                                   , iDestructVarIndex    = varIdx
                                                   , iDestructResultType  = varType
                                                   }
                , iExperimentArgs  = Seq.empty
                , iExperimentSubst = iExperimentSubst orig
                }
